import get
import compute
import output


def from_tcm_or_formula(tcm_or_formula,
                        score=990,
                        out_for_cytoscape=True,
                        out_graph=True,
                        re=True,
                        path='result/'):
    r"""
    进行经典的正向网络药理学分析
    :param tcm_or_formula: 任何可以使用in判断一个元素是否在其中的组合数据类型，拟分析的中药或复方的ID
    :param score: int类型，仅combined_score大于等于score的记录会被筛选出，默认为900
    :param out_for_cytoscape: 布尔类型，是否输出用于Cytoscape绘图的文件
    :param out_graph: 布尔类型，是否输出图
    :param re: 布尔类型，是否返回原始分析结果（中药、化合物（中药成分）、蛋白质（靶点）及其连接信息）
    :param path: 字符串类型，存放结果的目录
    :return: tcm: pd.DataFrame类型，中药信息
    :return: tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :return: chem: pd.DataFrame类型，化合物（中药成分）信息
    :return: chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :return: proteins: pd.DataFrame类型，蛋白质（靶点）信息
    """
    formula, tcm, formula_tcm_links = get.get_tcm_or_formula(tcm_or_formula)

    tcm_chem_links = get.get_tcm_chem_links('HVMID', tcm['HVMID'])
    chem = get.get_chemicals('HVCID', tcm_chem_links['HVCID'])
    chem_protein_links = get.get_chem_protein_links('HVCID', chem['HVCID'], score)
    proteins = get.get_proteins('Ensembl_ID', chem_protein_links['Ensembl_ID'])

    # 深度优先搜索筛选有效节点
    formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins = dfs_filter(
        formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins)

    tcm, chem, formula = compute.score(tcm, tcm_chem_links, chem, chem_protein_links, formula, formula_tcm_links)

    if out_for_cytoscape:
        output.out_for_cyto(tcm, tcm_chem_links, chem, chem_protein_links, proteins, path)

    if out_graph:
        output.vis(tcm, tcm_chem_links, chem, chem_protein_links, proteins, path)

    if re:
        if tcm_or_formula[0][2] == 'P':
            return formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins
        else:
            return tcm, tcm_chem_links, chem, chem_protein_links, proteins


def from_proteins(proteins,
                  score=0,
                  random_state=None,
                  num=1000,
                  tcm_component=True,
                  formula_component=True,
                  out_for_cytoscape=True,
                  re=True,
                  path='result/'):
    r"""
    进行逆向网络药理学分析
    :param proteins: 任何可以使用in判断一个元素是否在其中的组合数据类型，存储拟分析蛋白质（靶点）在STITCH中的Ensembl_ID
    :param score: int类型，仅combined_score大于等于score的记录会被筛选出，默认为0
    :param random_state: int类型，指定随机数种子
    :param num: int类型，指定优化时需生成的解的组数
    :param tcm_component: 布尔类型，是否优化中药组合
    :param formula_component: 布尔类型，是否优化复方组合
    :param out_for_cytoscape: 布尔类型，是否输出用于Cytoscape绘图的文件
    :param re: 布尔类型，是否返回原始分析结果
    :param path: 字符串类型，存放结果的目录
    :return: tcm: pd.DataFrame类型，中药信息
    :return: tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :return: chem: pd.DataFrame类型，化合物（中药成分）信息
    :return: chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :return: protein: pd.DataFrame类型，蛋白质（靶点）信息
    """
    proteins = get.get_proteins('Ensembl_ID', proteins)
    chem_protein_links = get.get_chem_protein_links('Ensembl_ID', proteins['Ensembl_ID'], score)
    chem = get.get_chemicals('HVCID', chem_protein_links['HVCID'])
    tcm_chem_links = get.get_tcm_chem_links('HVCID', chem['HVCID'])
    tcm = get.get_tcm('HVMID', tcm_chem_links['HVMID'])
    formula_tcm_links = get.get_formula_tcm_links('HVMID', tcm['HVMID'])
    formula = get.get_formula('HVPID', formula_tcm_links['HVPID'])

    # 深度优先搜索筛选有效节点
    formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins = dfs_filter(
        formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins)

    # 计算Score
    tcm, chem, formula = compute.score(tcm, tcm_chem_links, chem, chem_protein_links, formula, formula_tcm_links)

    # 优化模型
    tcms = compute.component(tcm.loc[tcm['Importance Score'] != 1.0], random_state, num) if tcm_component else None
    formulas = compute.component(formula.loc[formula['Importance Score'] != 1.0],
                                 random_state, num) if formula_component else None

    if out_for_cytoscape:
        output.out_for_cyto(tcm, tcm_chem_links, chem, chem_protein_links, proteins, path)

    if re:
        return formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins, tcms, formulas


def from_tcm_or_formula_proteins(tcm_or_formula,
                                 proteins,
                                 score=0,
                                 out_for_cytoscape=True,
                                 out_graph=True,
                                 re=True,
                                 path='result/'):
    r"""
    进行经典的正向网络药理学分析，并根据给定的靶点筛选结果
    :param tcm_or_formula: 任何可以使用in判断一个元素是否在其中的组合数据类型，拟分析的中药或复方的ID
    :param proteins: 任何可以使用in判断一个元素是否在其中的组合数据类型，拟分析蛋白质（靶点）在STITCH中的Ensembl_ID
    :param score: int类型，仅combined_score大于等于score的记录会被筛选出
    :param out_for_cytoscape: 布尔类型，是否输出用于Cytoscape绘图的文件
    :param out_graph: 布尔类型，是否输出图
    :param re: 布尔类型，是否返回原始分析结果
    :param path: 字符串类型，存放结果的目录
    :return: tcm: pd.DataFrame类型，中药信息
    :return: tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :return: chem: pd.DataFrame类型，化合物（中药成分）信息
    :return: chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :return: proteins: pd.DataFrame类型，蛋白质（靶点）信息
    """
    formula, tcm, formula_tcm_links = get.get_tcm_or_formula(tcm_or_formula)
    proteins = get.get_proteins('Ensembl_ID', proteins)

    # 根据中药和蛋白质（靶点）获取中药-化合物（中药成分）连接和化合物（中药成分）-蛋白质（靶点）连接
    tcm_chem_links = get.get_tcm_chem_links('HVMID', tcm['HVMID'])
    chem_protein_links = get.get_chem_protein_links('Ensembl_ID', proteins['Ensembl_ID'], score)

    # 最终得到的化合物必须与中药和蛋白均有连接，故此处仅get一次，选择中药是为了节省内存
    chem = get.get_chemicals('HVCID', tcm_chem_links['HVCID'])

    # 深度优先搜索筛选有效节点
    formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins = dfs_filter(
        formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins)

    # 计算Score
    tcm, chem, formula = compute.score(tcm, tcm_chem_links, chem, chem_protein_links, formula, formula_tcm_links)

    if out_for_cytoscape:
        output.out_for_cyto(tcm, tcm_chem_links, chem, chem_protein_links, proteins, path)

    if out_graph:
        output.vis(tcm, tcm_chem_links, chem, chem_protein_links, proteins, path)

    if re:
        if tcm_or_formula[0][2] == 'P':
            return formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins
        else:
            return tcm, tcm_chem_links, chem, chem_protein_links, proteins


def dfs_filter(formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins):
    """深度优先搜索筛选有效节点（在完整的（复方-）中药-化合物-蛋白通路中的节点）
    :param formula: pd.DataFrame类型，复方信息
    :param formula_tcm_links: pd.DataFrame类型，复方-中药连接信息
    :param tcm: pd.DataFrame类型，中药信息
    :param tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :param chem: pd.DataFrame类型，化合物（中药成分）信息
    :param chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :param proteins: pd.DataFrame类型，蛋白质（靶点）信息
    :return: formula: pd.DataFrame类型，复方信息
    :return: formula_tcm_links: pd.DataFrame类型，复方-中药连接信息
    :return: tcm: pd.DataFrame类型，中药信息
    :return: tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :return: chem: pd.DataFrame类型，化合物（中药成分）信息
    :return: chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :return: proteins: pd.DataFrame类型，蛋白质（靶点）信息
    """
    formula_id = set()
    tcm_id = set()
    chem_id = set()
    proteins_id = set()

    # 深度优先搜索得到有效节点的ID
    for f in formula['HVPID'] if formula_tcm_links is not None else [0]:
        for m in tcm['HVMID'] if formula_tcm_links is None else set(formula_tcm_links.loc[
                                                                        formula_tcm_links['HVPID'] == f]['HVMID']):
            for c in set(tcm_chem_links.loc[tcm_chem_links['HVMID'] == m]['HVCID']):
                for p in set(chem_protein_links.loc[chem_protein_links['HVCID'] == c]['Ensembl_ID']):
                    formula_id.add(f)
                    tcm_id.add(m)
                    chem_id.add(c)
                    proteins_id.add(p)

    # 根据有效节点的ID更新formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins
    formula = None if formula is None else formula.loc[formula['HVPID'].isin(formula_id)]
    tcm = tcm.loc[tcm['HVMID'].isin(tcm_id)]
    chem = chem.loc[chem['HVCID'].isin(chem_id)]
    proteins = proteins.loc[proteins['Ensembl_ID'].isin(proteins_id)]
    formula_tcm_links = None if formula_tcm_links is None else formula_tcm_links.loc[
        formula_tcm_links['HVPID'].isin(formula_id) & formula_tcm_links['HVMID'].isin(tcm_id)]
    tcm_chem_links = tcm_chem_links.loc[tcm_chem_links['HVMID'].isin(tcm_id) & tcm_chem_links['HVCID'].isin(chem_id)]
    chem_protein_links = chem_protein_links.loc[chem_protein_links['HVCID'].isin(chem_id) &
                                                chem_protein_links['Ensembl_ID'].isin(proteins_id)]

    # 重新编号（chem、tcm和formula在计算score时会重新编号，此处不再重新编号）
    tcm_chem_links.index = range(tcm_chem_links.shape[0])
    chem_protein_links.index = range(chem_protein_links.shape[0])
    proteins.index = range(proteins.shape[0])

    return formula, formula_tcm_links, tcm, tcm_chem_links, chem, chem_protein_links, proteins


if __name__ == '__main__':
    tcm_ft, tcm_chem_links_ft, chem_ft, chem_protein_links_ft, protein_ft = from_tcm_or_formula(['HVM0367', 'HVM1695'])
    formula_ff, formula_tcm_links_ff, tcm_ff, tcm_chem_links_ff, chem_ff, chem_protein_links_ff, protein_ff = \
        from_tcm_or_formula(['HVP1625'])
    formula_fg, tcm_fg, tcm_chem_l_fg, chem_fg, chem_protein_l_fg, protein_fg, tcms_fg, formulas_fg = from_proteins(
        ['ENSP00000381588', 'ENSP00000252519'], num=10)
    tcm_ftp, tcm_chem_links_ftp, chem_ftp, chem_protein_links_ftp, protein_ftp = \
        from_tcm_or_formula_proteins(['HVM0367', 'HVM1695'], ['ENSP00000381588', 'ENSP00000252519'])
    formula_ffp, formula_tcm_links_ffp, tcm_ffp, tcm_chem_links_ffp, chem_ffp, chem_protein_links_ffp, protein_ffp = \
        from_tcm_or_formula_proteins(['HVP1625'], ['ENSP00000381588', 'ENSP00000252519'])



import numpy as np
import pandas as pd
from math import ceil
import random
from tqdm import tqdm


def score(tcm,
          tcm_chem_links,
          chem,
          chem_protein_links,
          formula=None,
          formula_tcm_links=None):
    """
    计算化合物和中药的Importance Score
    :param formula_tcm_links:
    :param formula:
    :param tcm: pd.DataFrame类型，中药信息
    :param tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :param chem: pd.DataFrame类型，化合物（中药成分）信息
    :param chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :return: chem: pd.DataFrame类型，化合物的Importance Score的计算结果
    :return: tcm: pd.DataFrame类型，中药的Importance Score的计算结果
    """

    tcm_c = tcm.copy()
    chem_c = chem.copy()

    # 计算化合物的Importance Score
    chem_c['Importance Score'] = chem_c.loc[:, 'HVCID'].apply(
        lambda x: 1 - (
                1 - np.array([*chem_protein_links.loc[chem_protein_links['HVCID'] == x]['Combined_score']])).prod())

    # 计算中药的Importance Score
    tcm_c['Importance Score'] = tcm_c.loc[:, 'HVMID'].apply(
        lambda x: 1 - (1 - np.array([*chem_c.loc[
            chem_c['HVCID'].isin(tcm_chem_links.loc[tcm_chem_links['HVMID'] == x]['HVCID'])][
            'Importance Score']])).prod())

    # 若传入了复方相关信息，则需计算复方的Importance Score
    if formula is not None:
        formula_c = formula.copy()
        formula_c['Importance Score'] = formula_c.loc[:, 'HVPID'].apply(
            lambda x: 1 - (1 - np.array([*tcm_c.loc[
                tcm_c['HVMID'].isin(formula_tcm_links.loc[formula_tcm_links['HVPID'] == x]['HVMID'])][
                'Importance Score']])).prod())
        # 根据Importance Score降序排序
        formula_c = formula_c.sort_values(by='Importance Score', ascending=False)
        # 重新设置索引
        formula_c.index = range(formula_c.shape[0])
    else:
        formula_c = None

    # 根据Importance Score降序排序
    tcm_c = tcm_c.sort_values(by='Importance Score', ascending=False)

    # 重新设置索引
    tcm_c.index = range(tcm_c.shape[0])
    chem_c.index = range(chem_c.shape[0])

    return tcm_c, chem_c, formula_c


def component(items_and_score, random_state=None, num=1000, c=10):
    '''
    :param random_state:
    :param tcm:
    :param items_and_score: pd存储复方/中药信息
    :param num: 需要的解的组数
    :return:
    '''
    if 'HVPID' in items_and_score.columns:
        by = 'HVPID'
        name = 'name'
    else:
        by = 'HVMID'
        name = 'cn_name'

    dps = []
    items_ls = []
    n = len(items_and_score)
    weights = [1 for _ in range(n)]
    names = [*items_and_score.loc[:, by]]
    values = [*items_and_score.loc[:, 'Importance Score']]
    n = ceil(len(weights) / 10)

    if random_state is not None:
        random.seed(random_state)

    for _ in tqdm(range(num)):
        random_indices = random.sample(range(len(weights)), n)
        weights = [weights[i] for i in random_indices]
        names = [names[i] for i in random_indices]
        values = [values[i] for i in random_indices]

        # 不能再得出之前的解
        dp, items = knapsack(weights, n, items_ls, names, values, c)
        dps.append(dp)
        items_ls.append(items)

    # 用pd.DataFrame存储结果
    components = pd.DataFrame(dps)
    components.columns = ['Importance Score']
    components["items"] = items_ls
    components["items"] = components["items"].apply(lambda x: ';'.join(x))
    components = components.loc[:, ['items', 'Importance Score']]

    # 计算Score的提升量
    components['Boost'] = components.apply(boost, axis=1, args=(items_and_score, by))

    #components.loc[:, 'items'] = components.loc[:, 'items'].apply(
        #lambda x: ';'.join([str(items_and_score.loc[items_and_score[by] == x][name].iloc[0]) for x in x.split(';')]))

    # 根据boost降序排序
    components = components.sort_values(by='Boost', ascending=False)
    # 重新设置索引
    components.index = range(components.shape[0])

    return components


def boost(row, items_and_score, by):
    ls = row['items'].split(';')
    scores = [*items_and_score.loc[items_and_score[by].isin(ls)]['Importance Score']]
    return (row['Importance Score'] - max(scores))/max(scores)


def knapsack(weights, n, forbidden_combinations, names, values, c=10):
    # 创建一个二维列表用于存储计算结果
    dp = [[0] * (c + 1) for _ in range(n + 1)]
    # 创建一个二维列表用于记录选择的中药/复方
    items = [[""] * (c + 1) for _ in range(n + 1)]

    for i in range(1, n + 1):
        for j in range(1, c + 1):
            if weights[i - 1] <= j:
                # 检查当前是否与禁止组合冲突
                conflict = False
                for combination in forbidden_combinations:
                    if names[i - 1] in combination and any(
                            item in items[i - 1][j - weights[i - 1]] for item in combination):
                        conflict = True
                        break
                if conflict:
                    dp[i][j] = dp[i - 1][j]
                    items[i][j] = items[i - 1][j]
                else:
                    if 1 - (1 - values[i - 1]) * (1 - dp[i - 1][j - weights[i - 1]]) > dp[i - 1][j]:
                        dp[i][j] = 1 - (1 - values[i - 1]) * (1 - dp[i - 1][j - weights[i - 1]])
                        items[i][j] = names[i - 1] + ", " + items[i - 1][j - weights[i - 1]]
                    else:
                        dp[i][j] = dp[i - 1][j]
                        items[i][j] = items[i - 1][j]
            else:
                dp[i][j] = dp[i - 1][j]
                items[i][j] = items[i - 1][j]

    # 优化中药/复方存储
    items = [s[:-2] for s in items[-1]]

    # 计算累计Score比例
    score_ratio = np.cumsum(dp[-1]) / np.sum(dp)

    # 最大似然估计
    mle_estimates = (score_ratio - 1 / len(score_ratio)) / np.sqrt(2 / len(score_ratio))

    # 应选择的复方/中药数（即索引）但不能只选择一个
    num_components = np.argmin(mle_estimates) + 1
    num_components = 2 if num_components <= 1 else num_components

    return dp[-1][num_components], items[num_components].split(', ')


if __name__ == '__main__':
    import get

    protein_info = get.get_proteins('Ensembl_ID', ['ENSP00000381588', 'ENSP00000252519'])
    chem_protein_links_info = get.get_chem_protein_links('Ensembl_ID', protein_info['Ensembl_ID'], 0)
    chem_info = get.get_chemicals('HVCID', chem_protein_links_info['HVCID'])
    tcm_chem_links_info = get.get_tcm_chem_links('HVCID', chem_info['HVCID'])
    tcm_info = get.get_tcm('HVMID', tcm_chem_links_info['HVMID'])
    formula_tcm_links_info = get.get_formula_tcm_links('HVMID', tcm_info['HVMID'])
    formula_info = get.get_formula('HVPID', formula_tcm_links_info['HVPID'])

    tcm_info, chem_info, formula_info = score(tcm_info,
                                              tcm_chem_links_info,
                                              chem_info,
                                              chem_protein_links_info,
                                              formula_info,
                                              formula_tcm_links_info)
    tcms = component(tcm_info.loc[tcm_info['Importance Score'] != 1.0], random_state=138192, c=2)
    formulas = component(formula_info.loc[formula_info['Importance Score'] != 1.0], random_state=138192, c=5)


import os
import pandas as pd


def get_formula(by, items):
    """
    读取HerbiV_formula数据集，返回items中复方的信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: 任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的复方
    :return: pd.DataFrame类型，items中复方的信息
    """

    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    formula_all = pd.read_csv(current_directory + r'/data/HerbiV_formula.csv')

    # 在HerbiV_tcm中获取items中复方的信息
    formula = formula_all.loc[formula_all[by].isin(items)].copy()

    # 重新设置索引
    formula.index = range(formula.shape[0])

    return formula


def get_formula_tcm_links(by, items):
    """
    读取HerbiV_formula_tcm_links数据集，返回items中复方/中药的复方-中药连接信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: pd.DataFrame或其他任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的复方/中药
    :return: pd.DataFrame类型，items中复方/中药的复方-中药连接信息
    """
    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    formula_tcm_links_all = pd.read_csv(current_directory + r'/data/HerbiV_formula_tcm_links.csv')

    # 在HerbiV_formula_tcm_links中获取items中复方/中药的复方-中药连接信息
    formula_tcm_links = formula_tcm_links_all.loc[formula_tcm_links_all[by].isin(items)].copy()

    # 重新设置索引
    formula_tcm_links.index = range(formula_tcm_links.shape[0])

    return formula_tcm_links


def get_tcm(by, items):
    """
    读取HerbiV_tcm数据集，返回items中中药的信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: 任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的中药
    :return: pd.DataFrame类型，items中中药的信息
    """

    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    tcm_all = pd.read_csv(current_directory + r'/data/HerbiV_tcm.csv')

    # 在HerbiV_tcm中获取items中中药的信息
    tcm = tcm_all.loc[tcm_all[by].isin(items)].copy()

    # 重新设置索引
    tcm.index = range(tcm.shape[0])

    return tcm


def get_tcm_chem_links(by, items):
    """
    读取HerbiV_tcm_chemical_links数据集，返回items中中药/化合物的中药-成分信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: pd.DataFrame或其他任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的中药/化合物
    :return: pd.DataFrame类型，items中中药/化合物的中药-成分信息
    """
    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    tcm_chem_links_all = pd.read_csv(current_directory + r'/data/HerbiV_tcm_chemical_links.csv')

    # 在HerbiV_tcm_chemical_links中获取items中中药/化合物的中药-成分连接信息
    tcm_chem_links = tcm_chem_links_all.loc[tcm_chem_links_all[by].isin(items)].copy()

    # 重新设置索引
    tcm_chem_links.index = range(tcm_chem_links.shape[0])

    return tcm_chem_links


def get_chemicals(by, items):
    """
    读取HerbiV_chemicals数据集，返回items中化合物的信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: pd.DataFrame或其他任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的化合物
    :return: pd.DataFrame类型，items中化合物的信息
    """
    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    chem_all = pd.read_csv(current_directory + r'/data/HerbiV_chemicals.csv')

    # 在HerbiV_chemical_protein_links中获取items中化合物的信息
    chem = chem_all.loc[chem_all[by].isin(items)].drop_duplicates(subset=['HVCID'])

    # 重新设置索引
    chem.index = range(chem.shape[0])

    return chem


def get_chem_protein_links(by, items, score=900):
    """
    读取HerbiV_chemicals数据集，返回items中化合物/蛋白质的combined_score大于等于score的化合物-蛋白质（靶点）连接信息
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: pd.DataFrame或其他任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的化合物/蛋白质
    :param score: int类型，仅combined_score大于等于score的记录会被筛选出，默认为900
    :return: pd.DataFrame类型，items中化合物/蛋白质的combined_score大于等于score的化合物-蛋白质（靶点）连接信息
    """
    # 读取数据集
    current_directory = os.path.dirname(os.path.abspath(__file__))
    chem_protein_links_all = pd.read_csv(current_directory + r'/data/HerbiV_chemical_protein_links.csv')

    # 在HerbiV_chemical_protein_links中获取items中化合物/蛋白质的combined_score大于等于score的化合物-蛋白质（靶点）连接信息
    chem_protein_links = chem_protein_links_all.loc[
        (chem_protein_links_all[by].isin(items)) &
        (chem_protein_links_all['Combined_score'] >= score)].copy()

    # 将Combined_score变换为0-1的浮点数
    chem_protein_links.loc[:, 'Combined_score'] = chem_protein_links.loc[:, 'Combined_score'].apply(
        lambda x: x / 1000)

    # 重新设置索引
    chem_protein_links.index = range(chem_protein_links.shape[0])

    return chem_protein_links


def get_proteins(by, items):
    r"""
    从HerbiV_proteins中获得基因名与其对应的Ensemble_ID,从而可以被from_genes函数所调用，这样直接输入基因名即可
    :param by: str类型，数据集中与items相匹配的列的列名
    :param items: pd.DataFrame或其他任何可以使用in判断一个元素是否在其中的组合数据类型，存放要查询的蛋白质
    :return:返回一个包含对应信息的字典
    :return:返回一个数据框
    """

    # 数据的输入
    current_directory = os.path.dirname(os.path.abspath(__file__))
    proteins_all = pd.read_csv(current_directory + r'/data/HerbiV_proteins.csv')

    # 在HerbiV_chemical_proteins中获取items中化合物的信息
    proteins = proteins_all.loc[proteins_all[by].isin(items)].drop_duplicates(subset=['Ensembl_ID'])

    # 重置索引
    proteins.index = range(proteins.shape[0])

    return proteins


def get_tcm_or_formula(tcm_or_formula):
    """
    获取tcm_or_formula中元素对应的中药、复方及其连接信息
    :param tcm_or_formula: 任何可以使用in判断一个元素是否在其中的组合数据类型，拟获取的中药或复方的ID
    :return: formula: pd.DataFrame类型，复方信息
    :return: tcm: pd.DataFrame类型，中药信息
    :return: formula_tcm_links: pd.DataFrame类型，复方-中药连接信息
    """
    if tcm_or_formula[0][2] == 'P':  # 判断输入是否为复方的HVPID
        formula = get_formula('HVPID', tcm_or_formula)
        formula_tcm_links = get_formula_tcm_links('HVPID', formula['HVPID'])
        tcm = get_tcm('HVMID', formula_tcm_links['HVMID'])
    else:
        formula = None
        formula_tcm_links = None
        tcm = get_tcm('HVMID', tcm_or_formula)
    return formula, tcm, formula_tcm_links


if __name__ == '__main__':
    formula_info = get_formula('HVPID', ['HVP1625'])
    formula_tcm_links_info = get_formula_tcm_links('HVPID', formula_info['HVPID'])
    tcm_info = get_tcm('HVMID', formula_tcm_links_info['HVMID'])
    tcm_chem_links_info = get_tcm_chem_links('HVMID', tcm_info['HVMID'])
    chem_info = get_chemicals('HVCID', tcm_chem_links_info['HVCID'])
    chem_protein_links_info = get_chem_protein_links('HVCID', chem_info['HVCID'])
    protein_info = get_proteins('Ensembl_ID', chem_protein_links_info['Ensembl_ID'])
    formula_info1, tcm_info1, formula_tcm_links_info1 = get_tcm_or_formula(['HVP1625'])
    formula_info2, tcm_info2, formula_tcm_links_info2 = get_tcm_or_formula(['HVM0367', 'HVM1695'])



import os
import pandas as pd
from pyecharts import options as opts
from pyecharts.charts import Graph
from pyecharts.globals import ThemeType


def out_for_cyto(tcm, tcm_chem_links, chem, chem_protein_links, protein, path='result/'):
    r"""
    输出Cytoscape用于作图的网络文件和属性文件
    :param tcm: pd.DataFrame类型，中药信息
    :param tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :param chem: pd.DataFrame类型，化合物（中药成分）信息
    :param chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :param path: 字符串类型，存放结果的目录
    """
    tcm_c = tcm.copy()
    tcm_chem_links_c = tcm_chem_links.copy()
    chem_c = chem.copy()
    chem_protein_links_c = chem_protein_links.copy()
    protein_c = protein.copy()

    # 若无path目录，先创建该目录
    if not os.path.exists(path):
        os.mkdir(path)

    out_chem_protein_links = chem_protein_links_c.iloc[:, 0:2]
    out_chem_protein_links.columns = ['SourceNode', 'TargetNode']

    out_chem_protein_links.loc[:, 'SourceNode'] = out_chem_protein_links.loc[:, 'SourceNode'].apply(
        lambda x: chem_c.loc[chem_c['HVCID'] == x]['Name'].iloc[0] if len(
            chem_c.loc[chem_c['HVCID'] == x]['Name']) > 0 else None)

    out_chem_protein_links.dropna(subset=['SourceNode'], inplace=True)

    out_chem_protein_links.loc[:, 'TargetNode'] = out_chem_protein_links.loc[:, 'TargetNode'].apply(
        lambda x: protein_c.loc[protein_c['Ensembl_ID'] == x]['gene_name'].iloc[0] if len(
            protein_c.loc[protein_c['Ensembl_ID'] == x]['gene_name']) > 0 else None)

    out_chem_protein_links.dropna(subset=['TargetNode'], inplace=True)

    out_tcm_chem = tcm_chem_links_c.iloc[:, 0:2]
    out_tcm_chem.columns = ['SourceNode', 'TargetNode']

    out_tcm_chem.loc[:, 'SourceNode'] = out_tcm_chem.loc[:, 'SourceNode'].apply(
        lambda x: tcm_c.loc[tcm_c['HVMID'] == x]['cn_name'].iloc[0] if len(
            tcm_c.loc[tcm_c['HVMID'] == x]['cn_name']) > 0 else None)

    out_tcm_chem.dropna(subset=['SourceNode'], inplace=True)

    out_tcm_chem.loc[:, 'TargetNode'] = out_tcm_chem.loc[:, 'TargetNode'].apply(
        lambda x: chem_c.loc[chem_c['HVCID'] == x]['Name'].iloc[0] if len(
            chem_c.loc[chem_c['HVCID'] == x]['Name']) > 0 else None)

    out_tcm_chem.dropna(subset=['TargetNode'], inplace=True)

    out_chem = chem_c.loc[:, ['Name']]
    out_chem.columns = ['Key']
    out_chem['Attribute'] = 'Chemicals'

    out_tcm = tcm_c.loc[:, ['cn_name']]
    out_tcm.columns = ['Key']
    out_tcm['Attribute'] = 'TCM'

    out_gene = protein_c.loc[:, ['gene_name']]
    out_gene.columns = ['Key']
    out_gene['Attribute'] = 'Proteins'

    # 输出Network文件
    pd.concat([out_chem_protein_links, out_tcm_chem]).to_csv(path + 'Network.csv', index=False)

    # 输出Type文件
    pd.concat([out_tcm, out_chem, out_gene]).to_csv(path + 'Type.csv', index=False)


def vis(tcm, tcm_chem_links, chem, chem_protein_links, protein, path='result/'):
    r"""
    使用NetworkX可视化分析结果
    :param tcm: pd.DataFrame类型，中药信息
    :param tcm_chem_links: pd.DataFrame类型，中药-化合物（中药成分）连接信息
    :param chem: pd.DataFrame类型，化合物（中药成分）信息
    :param chem_protein_links: pd.DataFrame类型，化合物（中药成分）-蛋白质（靶点）连接信息
    :param protein: pd.DataFrame类型，蛋白质（靶点）连接信息
    :param path: 字符串类型，存放结果的目录
    """
    # 若无path目录，先创建该目录
    if not os.path.exists(path):
        os.mkdir(path)

    #chem_protein_links = pd.DataFrame(chem_protein_links)
    #tcm_chem_links = pd.DataFrame(tcm_chem_links)

    tcm_chem_merged_1 = pd.merge(tcm_chem_links, tcm, on='HVMID', how='left')
    tcm_chem_links['HVMID'] = tcm_chem_merged_1['pinyin_name']

    tcm_chem_merged_2 = pd.merge(tcm_chem_links, chem, on='HVCID', how='left')
    tcm_chem_links['HVCID'] = tcm_chem_merged_2['Name']

    tcm_chem_links = tcm_chem_links[['HVMID', 'HVCID']]

    chem_protein_merged_1 = pd.merge(chem_protein_links, chem, on='HVCID', how='right')
    chem_protein_links['HVCID'] = chem_protein_merged_1['Name']
    chem_protein_merged_2 = pd.merge(chem_protein_links, protein, on='Ensembl_ID', how='left')
    chem_protein_links['Ensembl_ID'] = chem_protein_merged_2['gene_name']

    chem_protein_links = chem_protein_links[['HVCID', 'Ensembl_ID']]

    tcm_chem_links = tcm_chem_links.rename(columns = {"HCMID": "pinyin_name", "HVCID": "Name"})
    chem_protein_links = chem_protein_links.rename(columns = {"HVCID": "Name", "Ensembl_ID": "gene_name"})

    nodes = []
    edges = []

    categories = [
        {"name": "中药", "color": "#61a0a8"},
        {"name": "化学成分", "color": "#f47920"},
        {"name": "靶点", "color": "#ca8622"},
    ]

    for index, row in tcm_chem_links.iloc[1:].iterrows():
        chinese_medicine = row[0]
        chemical_component = row[1]
        nodes.append({'name': chinese_medicine, "symbolSize": 20, 'category': 0, "color": "#1FA9E9"})
        nodes.append({'name': chemical_component, "symbolSize": 20, 'category': 1, "color": "#FFFF00"})
        edges.append({'source': chinese_medicine, 'target': chemical_component})

    for index, row in chem_protein_links.iloc[1:].iterrows():
        chemical_component = row[0]
        target = row[1]
        nodes.append({'name': chemical_component, "symbolSize": 20, 'category': 1, "color": "#FFFF00"})
        nodes.append({'name': target, "symbolSize": 20, 'category': 2, "color": "#000000"})
        edges.append({'source': chemical_component, 'target': target})
    unique_list = list(set(tuple(item.items()) for item in nodes))
    nodes = [dict(item) for item in unique_list]

    unique_list = list(set(tuple(item.items()) for item in nodes))
    nodes = [dict(item) for item in unique_list]

    Graph(init_opts=opts.InitOpts(width="2400px", height="1200px"))\
        .add(
        '',
        nodes=nodes,
        links=edges,
        categories=categories,
        repulsion=8000,
        layout="circular",
        is_rotate_label=True,
        linestyle_opts=opts.LineStyleOpts(color="source", curve=0.3),
        label_opts=opts.LabelOpts(position="right")
    )\
        .set_global_opts(
        title_opts=opts.TitleOpts(title=''),
        legend_opts=opts.LegendOpts(orient="vertical", pos_left="2%", pos_top="20%")
    )\
        .render(path=path + "Graph.html")



if __name__ == '__main__':
    import get

    formula_info = get.get_formula('HVPID', ['HVP1625'])
    formula_tcm_links_info = get.get_formula_tcm_links('HVPID', formula_info['HVPID'])
    tcm_info = get.get_tcm('HVMID', formula_tcm_links_info['HVMID'])
    tcm_chem_links_info = get.get_tcm_chem_links('HVMID', tcm_info['HVMID'])
    chem_info = get.get_chemicals('HVCID', tcm_chem_links_info['HVCID'])
    chem_protein_links_info = get.get_chem_protein_links('HVCID', chem_info['HVCID'], 990)
    protein_info = get.get_proteins('Ensembl_ID', chem_protein_links_info['Ensembl_ID'])

    chem_info = chem_info.loc[chem_info.loc[:, 'HVCID'].isin(chem_protein_links_info['HVCID'])]
    tcm_chem_links_info = tcm_chem_links_info.loc[tcm_chem_links_info.loc[:, 'HVCID'].isin(chem_info['HVCID'])]
    tcm_info = tcm_info.loc[tcm_info.loc[:, 'HVMID'].isin(tcm_chem_links_info['HVMID'])]
    # 重新编号（chem和tcm在计算score时会重新编号，此处不再重新编号）
    tcm_chem_links_info.index = range(tcm_chem_links_info.shape[0])

    vis(tcm_info, tcm_chem_links_info, chem_info, chem_protein_links_info, protein_info)
    out_for_cyto(tcm_info, tcm_chem_links_info, chem_info, chem_protein_links_info, protein_info)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="herbiv",
    version="0.1a8",
    description="HerbiV(Bidirectional and Visible Database of Herb)既是一个数据库，又是一个强大的数据分析平台，集成了50多万条方剂、中药、成分、靶点数据，以及经过检验的数据挖掘模型。HerbiV is far more than a database, which is also a powerful data analysis platform that integrates more than 500,000 prescriptions, traditional Chinese medicine, ingredients and target data. Moreover, two tested data mining models are contained in HerbiV.",
    # Optional
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MLi-lab-Bioinformatics-NJUCM/HerbiV",
    author="王皓阳",
    author_email="Wesady@qq.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Natural Language :: Chinese (Simplified)",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ],
    keywords="network pharmacology",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.8",
    install_requires=[
                      "numpy",
                      "pandas",
                      "tqdm",
                      "pyecharts"
                      ],
    project_urls={
        "Bug Reports": "https://github.com/MLi-lab-Bioinformatics-NJUCM/pharmastar/issues",
        "Source": "https://github.com/MLi-lab-Bioinformatics-NJUCM/pharmastar",
    },
)
